#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "object.h"
#include "progstub.h"

// Names of variables defined by linker
int   SymCmp(const void *e1, const void *e2)
{
   return (((SymRef *)e1)->ofs - ((SymRef *)e2)->ofs);
}

int   AStrnCmp(const char *s1, const char *s2, int n)
{
   while (n--)
	if (tolower(*(s1++)) != tolower(*(s2++)))
	    return 1;
   return 0;
}

Object::Object(char *name)
{
   // Initialize information variables
   removeExt(name, filename);
   output89=0; output92=0; output92Plus=0;
   restoreScreen = 1; callExit = 0; readOnly = 0;
   mainOfs=OfsNotDefined; commentOfs=OfsNotDefined; exitOfs=OfsNotDefined;
   extraRAMOfs=OfsNotDefined; stubType=StubNormal;
   code=NULL; codeSize=0; bssSize=0;
   reloc[RelocCode].count=0; reloc[RelocCode].offset = 0; reloc[RelocCode].ofs=NULL;
   reloc[RelocBSS].count=0;  reloc[RelocBSS].offset = 0;  reloc[RelocBSS].ofs=NULL;
   reloc[RelocData].count=0; reloc[RelocData].offset = 0; reloc[RelocData].ofs=NULL;
   exportCount=0; ROM.count=0; RAM.count=0; libCount=0; special.count=0;
   version = 0; Debug = 0; 
   DSymCount = 0, SymCount = 0;
   StandardExtramRamCall = 0;
}

Object::~Object()
{
    for(int i = 0 ; i < 3 ; i++)
	free(reloc[i].ofs);
}

int Object::Read(char *fn)
{
   char buffer[100], file[100];
   removeExt(fn, buffer);
   sprintf(file,"%s.o",buffer);
   FILE *fp=fopen(file,"rb");
   if (fp == NULL)
      ERROR1("Cannot open Object file: %s\n",file);

   // Process each hunk in the object file
   while (!feof(fp))
   	{
	int hunk=ReadDWord(fp);
	EXTRA_INFO(printf("Hunk: %d\n", hunk));
	// Merge code & data ?
	if (hunk==HunkCode)
		{
		int size=ReadDWord(fp); // Size is in DWORDs
		// Allocate space for the code
		reloc[RelocCode].offset = codeSize;
		codeSize += size<<2;
		code = (char *) realloc(code, codeSize);
		EXTRA_INFO(printf(">Code Size = 0x%X\n", codeSize));
		// Read the code into memory
		fread(code,size,4,fp);
		}
	else if (hunk==HunkBSS)
		{
		reloc[RelocBSS].offset = bssSize;
		bssSize+=ReadDWord(fp)<<2;
		EXTRA_INFO(printf(">BSS Size = 0x%X\n", bssSize));
		}
	else if (hunk==HunkR32)
		{
		// Read each section's relocations
		while (1)
			{
			int count=ReadDWord(fp);
			if (!count) break;
			int section=ReadDWord(fp);
			if ((section==RelocCode)||(section==RelocBSS))
				{
				int i = reloc[section].count;		// Start
				reloc[section].count+=count;		// Number of symbols
				reloc[section].ofs= (int *) realloc(reloc[section].ofs, reloc[section].count*4); // Realloc table
				// Read in the relocation info
				for(int j = 0 ; j < count; i++, j++)
					{
					reloc[section].ofs[i]=ReadDWord(fp)+reloc[RelocCode].offset;
					AddDWord(code + reloc[section].ofs[i], reloc[section].offset);
					}
				}
			else	{
				printf("! Relocation of an unkwown section %u\n", section);
				for (int i=0;i<count;i++)
					ReadDWord(fp);
				}
			}
		}
	else if (hunk==HunkR16 || hunk==HunkR8)
		{
		int count=ReadDWord(fp);
		int section=ReadDWord(fp);
		ERROR2("Error: 16-bit & 8-bit relocations are unsupported\n"
                	"Section: %u Count: %u\n", section, count);
		}
	else if (hunk==HunkExt) // Exports or external references
		{
		// Read each entry
		while (1)
			{
			int size=ReadDWord(fp);
			// High byte of size contains type byte
			int type=(size>>24)&0xff;
			size&=0xffffff;
			if (type==TypeEnd)
				break;
			// Read name of symbol
			char sym[(size<<2)+1];
			int i;
			for (i=0;i<(size<<2);i++)
				sym[i]=fgetc(fp);
			sym[i]=0;
			EXTRA_INFO(printf("Exported symbol : %s\n", sym));
			if (type==TypeExport)
				{
				// Get offset of exported function
				int ofs=ReadDWord(fp)+reloc[RelocCode].offset;
				// Check for special export names
				if (!strcmp(sym,"_ti89"))
					output89=1;
				else if (!strcmp(sym,"_ti92"))
					output92=1;
				else if (!strcmp(sym,"_ti92plus"))
					output92Plus=1;
				else if (!strcmp(sym,"_v200"))
					outputV200=1;
				else if (!strcmp(sym,"_nosavescreen"))
					restoreScreen = 0;
				else if (!strcmp(sym,"_main"))
					mainOfs=ofs;
				else if (!strcmp(sym,"_comment"))
					commentOfs=ofs;
				else if (!strcmp(sym,"_EXIT"))
					exitOfs=ofs, callExit = 1;
				else if (!strcmp(sym,"_exit"))
					exitOfs=ofs;
			        else if (!strcmp(sym,"_readonly"))
					readOnly = 1;
				else if (!strcmp(sym,"_debug"))
					Debug = 1;
				else if (!strcmp(sym,"_extraram"))
					{
					if (StandardExtramRamCall)
						ERROR("ERROR: Can't mix standard standard ExtraRamCalls and IS89_xx_yy ramcalls !\n");
					StandardExtramRamCall = 1; // Doors way
					extraRAMOfs=ofs;
					}
				else if (!strcmp(sym,"_nostub"))
					stubType=StubNone;
				else if (!strcmp(sym,"_mistub"))
					stubType=StubMedium;
				else if (!strcmp(sym,"_install_preos"))
					stubType=StubInstall;
				else if (!strcmp(sym,"_library"))
					stubType=StubLibrary;
			        else if (!AStrnCmp(sym,"_version",8))
			        	{
			        	version = 1000;
		        		sscanf(&sym[strlen(sym)-2],"%x",&version);
					printf("Program Version: $%X\n", version);
					if (version > 255 )
		              			ERROR("Error: The version of the program couldn't be > 255");
					}
				else if (strlen(sym)>10 && !AStrnCmp(sym+strlen(sym)-10,"@version",8))
					{
		                 	int num=1000; // Initialize in case sscanf fails
				 	sscanf(&sym[strlen(sym)-2],"%x",&num);
		                     	printf("LIB %s version $%X \n", sym, num);
					if (num > 255)
		              			ERROR("Error: Version Number > 255");
					sym[strlen(sym)-10]=0;
					lib[AddLib(sym)].version = num;
		                     	}
				else if (!AStrnCmp(filename, sym, strlen(filename)) && sym[strlen(sym)-5]=='@')
					{
					// Check to make sure number is correct
					int num=-1; // Initialize in case sscanf fails
					sscanf(&sym[strlen(sym)-4],"%x",&num);
					if (num!=exportCount)
						ERROR("Error: Exported function xdef out of order or invalid.\n"
							"You must define the exported functions in sequential order !\n");
					// Add label to export table
					exportOfs[exportCount++]=ofs;
					}
				else	printf("Warning: Exporting %s is useless\n", sym);
				}
			else if (type==TypeImport32)
				{
				// Read number of references
				int count=ReadDWord(fp);
				if (!strncmp(sym,"_ROM_CALL_",10))
					{
					// Add references to ROM reference table
					int func=-1; // Initialize in case sscanf fails
					sscanf(&sym[10],"%x",&func);
					if (func==-1)
						ERROR1("Error: cannot read rom call value '%s'\n",sym);
					for (i=0;i<count;i++)
						{
						ROM.ref[ROM.count].type=LibRefDWord;
						ROM.ref[ROM.count].func=func;
						ROM.ref[ROM.count++].ofs=ReadDWord(fp)+reloc[RelocCode].offset;
						}
					}
				else if (!strncmp(sym,"_RAM_CALL_",10))
					{
					// Add references to RAM reference table
					int func=-1; // Initialize in case sscanf fails
					sscanf(&sym[10],"%x",&func);
					if (func==-1)
						ERROR1("Error: cannot read ram call value '%s'\n",sym);
					for (i=0;i<count;i++)
						{
						RAM.ref[RAM.count].type=LibRefDWord;
						RAM.ref[RAM.count].func=func;
						RAM.ref[RAM.count++].ofs=ReadDWord(fp)+reloc[RelocCode].offset;
						}
					}
				else if (!strncmp(sym,"_extraramaddr@",14))
					{
					// Add references to RAM reference table, setting the extra flag
					int func=-1; // Initialize in case sscanf fails
					sscanf(&sym[14],"%x",&func);
					if (func==-1)
						ERROR1("Error: cannot read extra ram call value '%s'\n",sym);
					for (i=0;i<count;i++)
						{
						RAM.ref[RAM.count].type=LibRefDWord|LibRefExtra;
						RAM.ref[RAM.count].func=func;
						RAM.ref[RAM.count++].ofs=ReadDWord(fp)+reloc[RelocCode].offset;
						}
					}
				else if (strlen(sym)>=5 && sym[strlen(sym)-5]=='@') // Library reference
					{
					// Add library reference to table
					int func=-1; // Initialize in case sscanf fails
					sscanf(&sym[strlen(sym)-4],"%x",&func);
					if (func==-1)
						ERROR1("Error: Undefined symbol '%s'\n",sym);
					sym[strlen(sym)-5]=0; // Strip number from name
					int libNum=AddLib(sym);
					for (i=0;i<count;i++)
						{
						lib[libNum].ref[lib[libNum].count].type=LibRefDWord;
						lib[libNum].ref[lib[libNum].count].func=func;
						lib[libNum].ref[lib[libNum].count++].ofs=ReadDWord(fp)+reloc[RelocCode].offset;
						}
					}
		               else // Unknow Imported labels
					{
					printf("WARNING !!! WARNING !!! TRY TO IMPORT AN UNKNOW LABEL : %s\n", sym);
					strncpy(Symbol[SymCount].name, sym, 128);
					Symbol[SymCount].ofs =ReadDWord(fp);
					SymCount++;
					}
				}
			else if ((type==TypeImport16)||(type==TypeImport16Alt))
				{
				// Read number of references
				int count=ReadDWord(fp);
				if (!strncmp(sym,"_RAM_CALL_",10))
					{
					// Add references to RAM reference table
					int func=-1; // Initialize in case sscanf fails
					sscanf(&sym[10],"%x",&func);
					if (func == -1)
						ERROR1("Symbol %s is not a valid RAM_CALL\n",sym);
					for (i=0;i<count;i++)
						{
						RAM.ref[RAM.count].type=LibRefWord;
						RAM.ref[RAM.count].func=func;
						RAM.ref[RAM.count++].ofs=ReadDWord(fp)+reloc[RelocCode].offset;
						}
					}
				else if (!strncmp(sym,"_extraramaddr@",14))
					{
					// Add references to RAM reference table, setting the extra flag
					int func=-1; // Initialize in case sscanf fails
					sscanf(&sym[14],"%x",&func);
					if (func == -1)
						ERROR1("Symbol %s is not a valid EXTRA_RAM_CALL\n",sym);
					for (i=0;i<count;i++)
						{
						RAM.ref[RAM.count].type=LibRefWord|LibRefExtra;
						RAM.ref[RAM.count].func=func;
						RAM.ref[RAM.count++].ofs=ReadDWord(fp)+reloc[RelocCode].offset;
						}
					}
				else	ERROR1("Sorry, cannot import symbol %s as 16-bits.\n",sym);
				}
			else if (type==TypeImport8)
				ERROR1("Error: 8-bit imports are unsupported for symbol '%s'\n",sym)
			else 
				ERROR2("Error: Unknown symbol '%s' type number $%2.2X\n",sym,type);
			}
		}
	else if (hunk==HunkEnd) // End of code or BSS section
		{
			// Do nothing
		}
	else if (hunk==HunkSym)
		{
		int size;
		int i;
		while ((size = ReadDWord(fp)) != 0)
			{
			// Read name of symbol
			for (i=0;i<(size<<2);i++)
				DSymbol[DSymCount].name[i]=fgetc(fp);
			DSymbol[DSymCount].name[i]=0;
			DSymbol[DSymCount].ofs = ReadDWord(fp)+reloc[RelocCode].offset;
			EXTRA_INFO(printf("Symbol %s Offset %X\n", DSymbol[DSymCount].name, DSymbol[DSymCount].ofs));
			DSymCount++;
			}
		qsort(DSymbol, DSymCount, sizeof(SymRef), SymCmp);
		}
	else if (!feof(fp))
		{
		int size=ReadDWord(fp); // Size is in DWORDs
		// Unused hunk, so skip it
	        EXTRA_INFO(printf(">Unused section\n"));
		fseek(fp,size<<2,SEEK_CUR);
		}
	}

   // Fix the unused libs bugs...
   for(int i = 0 ; i < libCount ;)
           {
           if (lib[i].count == 0)
              {
              for(int j = i ; j < libCount; j++)
                      lib[j] = lib[j+1];
              libCount--;
              }
           else i++;
           }

   printf("+OBJECT:%s CODE SIZE:%d BSS SIZE:%d \n RELOC CODE:%d RELOC BSS:%d ROM CALLS:%d RAM CALLS:%d LIBRARIES:%d SYMBOLS:%d\n",
                 fn, codeSize, bssSize, reloc[0].count, reloc[1].count, ROM.count, RAM.count, libCount, DSymCount);
   return 0;
}


int Object::AddLib(char *name)
{
	int found=-1;

	// Search library table for name
	for (int i=0;i<libCount;i++)
	{
		if (!strcmp(name,lib[i].name))
		{
			found=i;
			break;
		}
	}
	if (found==-1)
	{
		// Add a new library
		strcpy(lib[libCount].name,name);
		lib[libCount].version = 0;
		lib[libCount++].count = 0;
		return libCount-1;
	}
	else
		return found;
}

void Object::WriteData()
{
  FILE *out=fopen(filename,"wb");
  int i,csum;
  // Write signature
  if (calc==89)
	{
	fprintf(out,"**TI89**");
	printf("+TARGET: TI-89  SIZE: %d\n", size+3);
	}
  else
	{
	fprintf(out,"**TI92P*");
	printf("+TARGET: TI-92+ SIZE: %d\n", size+3);
	}
   fputc(1,out); fputc(0,out); // Unknown
   // Output folder name
   fprintf(out,"main");
   fputc(0,out); fputc(0x8b,out); fputc(0x76,out); fputc(0xa,out);
   // File comment (doesn't seem to work if not filled with zero bytes)
   for (i=0;i<40;i++)
	fputc(0,out);
   fputc(1,out); fputc(0,out); fputc(0x52,out); fputc(0,out); // Unknown
   fputc(0,out); fputc(0,out); // Unknown
   // Output program name
   for(i=0;i<8;i++)
	{
	if (filename[i]=='.') break;
	fputc(filename[i],out);
	}
   for( ;i<8;i++)
	fputc(0,out);
   // Output program type
   fputc(0x21,out); fputc(0,out); fputc(0,out); fputc(0,out);
   // Output file length
   fputc((size+0x5b)&0xff,out);
   fputc((size+0x5b)>>8,out);
   fputc(0,out); fputc(0,out);
   fputc(0xa5,out); fputc(0x5a,out); fputc(0,out); fputc(0,out); // Unknown
   fputc(0,out); fputc(0,out); // Unknown
   // Output program length
   fputc((size+1)>>8,out);
   fputc((size+1)&0xff,out);
   // Add program length to checksum
   csum=(size+1)&0xff;
   csum+=((size+1)>>8)&0xff;
   // Output program data
   for (i=0;i<size;i++)
	{
	fputc(data[i],out);
	csum+=data[i]&0xff;
	}
   fputc(0xf3,out); // Assembly program token
   csum+=0xf3;
   // Output checksum
   fputc(csum&0xff,out);
   fputc((csum>>8)&0xff,out);
   fclose(out);
}

void Object::GetOffsets(int &bssTableOfs,int &exportTableOfs,int &stubCodeOfs,int &stubLen)
{
   stubLen=0x1a; // 0x1a is the start of the lib relocation table, the first
		// dynamically sized section of the program header
   // Library import table
   stubLen+=2;            // Number of libs used
   stubLen+=libCount*10;  // Library names
   // Library relocation info
   for (int i=0;i<libCount;i++)
	{
	stubLen+=2; // Number of functions used in lib - 1
	int func=-1;
	// Examine library functions used
	for (int j=0;j<lib[i].count;j++)
		{
		if (func!=lib[i].ref[j].func)
			{
			if (func!=-1) stubLen+=2; // Terminate old list if needed
			func=lib[i].ref[j].func;
			stubLen+=2; // Function number
			}
		stubLen+=2; // Relocation offset
		}
	stubLen+=2;    // End the list from the last function
	}
   // ROM reference table
   stubLen+=2;      // Number of ROM references
   // Output ROM relocation info
   int func=-1;
   if (ROM.count) // Table exists only if there are ROM references
	{
	stubLen+=2; // Number of ROM functions used - 1
	// Examine ROM functions used
	for (int j=0;j<ROM.count;j++)
		{
		if (func!=ROM.ref[j].func)
			{
			if (func!=-1) stubLen+=2; // Terminate old list if needed
			func=ROM.ref[j].func;
			stubLen+=2; // Function number
			}
		stubLen+=2; // Relocation offset
   		}
	stubLen+=2; // End the list from the last function
	}
   // RAM reference table
   stubLen+=2; // Number of RAM references
   // Output RAM relocation info
   func=-1;
   if (RAM.count) // Table exists only if there are RAM references
	{
	stubLen+=2; // Number of RAM functions used - 1
	// Examine RAM functions used
	for (int j=0;j<RAM.count;j++)
		{
		if (func!=(RAM.ref[j].func|RAM.ref[j].type))
			{
			if (func!=-1) stubLen+=2; // Terminate old list if needed
			func=RAM.ref[j].func|RAM.ref[j].type;
			stubLen+=2;             // Function number
			}
		stubLen+=2; // Relocation offset
		}
	stubLen+=2; // End the list from the last function
	}
   // Code relocation table
   stubLen+=2+(reloc[0].count*2); // Relocations and terminating marker
   // BSS relocation table
   bssTableOfs=stubLen;
   stubLen+=6+(reloc[1].count*2); // Size, relocations and terminating marker
   // Export table
   exportTableOfs=stubLen;
   stubLen+=4+(exportCount*2); // Count, offsets, and terminating marker
   stubCodeOfs=stubLen;
   // Stub loader code
   if (stubType == StubMedium)
	stubLen+=progMiStubLen;
   else if (stubType == StubInstall)
	stubLen+=progInstallStubLen;
   else if (stubType!=StubLibrary)
	stubLen+=progStubLen; 

   //printf("+STUB SIZE: %d\n", stubLen);
}

int Object::Write(int _calc)
{
   calc = _calc;
   // Override old file name
   char buffer[100];
   removeExt(filename, buffer);
   sprintf(filename,"%s.%sz",buffer,(calc==89)?"89":"9x");
   
   ResetData();

	// Fill in linker-defined references with their values
	/*for (int i=0;i<special.count;i++)
	{
		if (special.ref[i].type&LibRefWord)
			AddWord(&code[special.ref[i].ofs],PlusShellSpecialVal[special.ref[i].func]);
		else // Dword reference
			AddDWord(&code[special.ref[i].ofs],PlusShellSpecialVal[special.ref[i].func]);
	}*/

   if (stubType != StubNone)
	{
	// Get size of stub and offset of tables
	int bssTableOfs,exportTableOfs,stubCodeOfs,stubLen;
	GetOffsets(bssTableOfs,exportTableOfs,stubCodeOfs,stubLen);	
	if (stubType==StubLibrary)
		{
		// Libraries cannot be executed, so first instruction is rts
		WriteWord(0x4e75);
		WriteWord(0x4e75);
		// Write signature dword
		WriteByte('6'); WriteByte('8'); WriteByte('k'); WriteByte('L');
		}
	else // File is a program
		{	
		// Branch to stub loader code
		WriteDWord(0x61000000+(stubCodeOfs-2));
		// Write signature dword
		WriteByte('6'); WriteByte('8'); WriteByte('k'); WriteByte('P');
		}

   // Extra Flags
   WriteByte(0);
   // Relocation count - always zero at first
   WriteByte(0);
   // Offsets to _main, _comment, and _exit
   if (commentOfs==OfsNotDefined)
	WriteWord(0); // No _comment
   else	WriteWord(commentOfs+stubLen);
   if (mainOfs==OfsNotDefined)
	WriteWord(0); // No _main
   else WriteWord(mainOfs+stubLen);
   if (exitOfs==OfsNotDefined)
	WriteWord(0); // No _exit
   else WriteWord(exitOfs+stubLen);
   // Version Number
   WriteByte(version);
   // Compatibility flags
   WriteByte((output92Plus?1:0)
               |(output89?2:0)
               |(restoreScreen?0:4)
               |(readOnly?8:0)
               |(output92?16:0)
               |(outputV200?32:0));
   // BSS handle - used by kernel when program is run
   WriteWord(0);
   // Offset to BSS, export, and extra RAM tables
   if (bssSize)
	WriteWord(bssTableOfs);
   else WriteWord(0);
   WriteWord(exportTableOfs);
   if (extraRAMOfs==OfsNotDefined)
	WriteWord(0); // No extra RAM table
   else	WriteWord(extraRAMOfs+stubLen);
   // Library import table
   WriteWord(libCount); // Number of libs used
   // Output names of all libraries
   for (int i=0;i<libCount;i++)
	{
	int j;
	for (j=0;j<8;j++)
		{
		if (!lib[i].name[j]) break;
		WriteByte(lib[i].name[j]);
		}
	for (;j<9;j++)
		WriteByte(0);
        WriteByte(lib[i].version);
	}
   // Output library relocation info
   for (int i=0;i<libCount;i++)
	{
	int count=0;
	int func=-1;
	// Count the number of library functions used
	for (int j=0;j<lib[i].count;j++)
		{
		if (func!=lib[i].ref[j].func)
			{
			count++;
			func=lib[i].ref[j].func;
			}
		}
	// Add stub size to lib relocation offsets
	for (int j=0;j<lib[i].count;j++)
		lib[i].ref[j].ofs+=stubLen;
	WriteWord(count-1);           // Number of functions used in lib - 1
	// Output table
	func=-1;
	for (int j=0;j<lib[i].count;j++)
		{
		if (func!=lib[i].ref[j].func)
			{
			if (func!=-1) WriteWord(0); // Terminate old list if needed
			func=lib[i].ref[j].func;
			WriteWord(func);				
			}
		WriteWord(lib[i].ref[j].ofs);
		}
	// End the list from the last function
	WriteWord(0);
	}

    // ROM reference table
    WriteWord(ROM.count);
    if (ROM.count) // Table only exists if there are ROM references
	{
	// Output ROM relocation info
	int count=0;
	int func=-1;
	// Count the number of ROM functions used
	for (int j=0;j<ROM.count;j++)
		{
		if (func!=ROM.ref[j].func)
			{
			count++;
			func=ROM.ref[j].func;
			}
		}
	// Add stub size to ROM relocation offsets
	for (int j=0;j<ROM.count;j++)
		ROM.ref[j].ofs+=stubLen;
	WriteWord(count-1); // Number of ROM functions used - 1
	// Output table
	func=-1;
	for (int j=0;j<ROM.count;j++)
		{
		if (func!=ROM.ref[j].func)
			{
			if (func!=-1) WriteWord(0); // Terminate old list if needed
			func=ROM.ref[j].func;
			WriteWord(func);				
			}
		WriteWord(ROM.ref[j].ofs);
		}
	// End the list from the last function
	WriteWord(0);
	}

    // RAM reference table
    WriteWord(RAM.count);
    if (RAM.count) // Table only exists if there are RAM references
	{
	// Output RAM relocation info
	int count=0;
	int func=-1;
	// Count the number of RAM functions used
	for (int j=0;j<RAM.count;j++)
		{
		if (func!=(RAM.ref[j].func|RAM.ref[j].type))
			{
			count++;
			// Function number and type are combined into one value
			// to save space
			func=RAM.ref[j].func|RAM.ref[j].type;
			}
		}
	// Add stub size to RAM relocation offsets
	for (int j=0;j<RAM.count;j++)
		RAM.ref[j].ofs+=stubLen;
	WriteWord(count-1); // Number of RAM functions used - 1
	// Output table
	func=-1;
	for (int j=0;j<RAM.count;j++)
		{
		if (func!=(RAM.ref[j].func|RAM.ref[j].type))
			{
			if (func!=-1) WriteWord(0); // Terminate old list if needed
			func=RAM.ref[j].func|RAM.ref[j].type;
			WriteWord(func);				
			}
		WriteWord(RAM.ref[j].ofs);
		}
	// End the list from the last function
	WriteWord(0);
	}
    // Code relocation table
    for (int i=0;i<reloc[0].count;i++)
	{
	// Update destination offset in code section
	AddDWord(&code[reloc[0].ofs[i]],stubLen);
	// Write relocation offset
	WriteWord(reloc[0].ofs[i]+stubLen);
	}
    WriteWord(0); // End relocation table
    // BSS relocation table
    WriteDWord(bssSize);
    for (int i=0;i<reloc[1].count;i++)
	WriteWord(reloc[1].ofs[i]+stubLen);
    WriteWord(0); // End relocation table
    // Export table
    WriteWord(exportCount);
    for (int i=0;i<exportCount;i++)
	WriteWord(exportOfs[i]+stubLen);
    WriteWord(0); // End export table
    // Output stub code
    if (stubType!=StubLibrary)
	{
	if (stubType == StubMedium)
		{
		for (int i=0;i<progMiStubLen;i++)
			WriteByte(progMiStub[i]);
		}
	else if (stubType == StubInstall)
		{
		for (int i=0;i<progInstallStubLen;i++)
			WriteByte(progInstallStub[i]);
		}
	else
		for (int i=0;i<progStubLen;i++)
			WriteByte(progStub[i]);
	}
   // Output code
   for (int i=0;i<codeSize;i++)
	WriteByte(code[i]);
   // Debug infos
   if (Debug)
	{
	WriteByte(0);
	WriteWord(0xFFFF);   // Symbol Table End
	for(int i = 0 ; i < DSymCount ; i++)
		{
		for(int j = 0 ; DSymbol[i].name[j] != 0 ; j++)
			WriteByte(DSymbol[i].name[j]);
		WriteByte(0); // String ends
		WriteWord(DSymbol[i].ofs + stubLen);
		}
	}

   WriteWord(0); // Termination marker for TI's relocation method
   } 
   else 	// No stub
	{
	// Output code
	for (int i=0;i<codeSize;i++)
		WriteByte(code[i]);
	// Code relocation table
	for (int i=0;i<reloc[0].count;i++)
		{
		WriteWord(ReadDWord(reloc[0].ofs[i]));
		WriteWord(reloc[0].ofs[i]);
		}
	WriteWord(0); // Termination marker for TI's relocation method
	}
  WriteData();
  return 0;
}
