; Kernel Reloc file
;
;	It needs also many functions from ext.asm.
;	And you also defined some variables (FirstRun / Error strings / OrgStack / ProgramRef)
;
; NOTE: It used a lot SymFindFirst/SymFindNext, so if you have many files, it could be slow.

RELOC_MAC	MACRO
	tst.b	d7
	beq.s	\\@reloc
		neg.l	\1
\\@reloc
		ENDM

; Called by the relocation function when there is an error.
; Avoid using ER_catch by doing ourselves.
; In :
;	d0.w = Error number
;	a3 -> Library Name
relocation::error
	move.l	RelocStackList-FirstRun(a6),a7	; Restore Stack
	move.l	4(a7),a5			; Restore Main Program Ref

	tst.b	d7				; If we are relocating
	bne.s	\no_unrealloc			; Now unreloc the program before processing the error !
		moveq	#1,d7			; Unreloc the program.
		bsr	relocation_recur	; It will return here with the same exeption !
\no_unrealloc:
	; Get the string of the error		
	ifnd	PEDROM				
		lea	mlibrairie(pc),a1	; Complete the string in case of 
		move.l	(a3)+,(a1)+		; a missing library
		move.l	(a3),(a1)+
		lea	mmauvaise+4(pc),a1	; Complete the string in case of 
		move.l	(a3),(a1)		; a wrong library
		move.l	-(a3),-(a1)
	endif
	ifd	FLASH_CODE			; If the target is in Flash, we can't write directly since it is Write-protected.
		lea	KERNEL_EXTRA_STR-FirstRun(a6),a1	; Save the name of the missing library
		move.l	(a3)+,(a1)+		; For Pedrom
		move.l	(a3)+,(a1)+
		clr.w	(a1)+
	endif	
	add.w	d0,d0				; Number *2
	move.w	Error_Table(Pc,d0.w),d0		; Get the offset
	lea	Error_Table(Pc,d0.w),a0		; Get the error string
	move.l	a0,ErrorString-FirstRun(a6)	; Save the error string

	; Find the archived library which are not cleared.
	; ie with reloc_count = 0, locked !
	; I don't know how to avoid this case in my reloc process.
\start	move.w	#2,-(a7)
	clr.l	-(a7)
	ROM_THROW SymFindFirst
	addq.l	#6,a7
\loop						; Loop the VAT
		bsr	IsKernelPrgm		; Test if it is a Kernel program.
		tst.w	d0
		beq.s	\next
			move.w	SYM_ENTRY.hVal(a0),-(a7)
			ROM_THROW HeapDeref	; Get the address
			addq.l	#2,a7
			lea	2(a0),a5	; Get program address
			tst.b	9(a5)		; Reloc count = 0?
			bne.s	\next		; No.
				cmp.l	#$200000,a0			; Archived ?
				bhi.s	\next				; Yes so skip
					btst.b	#7,-2(a0)		; Locked Handle ?
					beq.s	\next			; No so skip
						bsr	reloc_free	; Unlock the handle and free the copy if it exists
						bra.s	\start		; ReStart the file check (Because Reloc_free uses also SymFindFirst/SymFindNext)
\next		ROM_THROW SymFindNext
		move.l	a0,d0
		bne.s	\loop

	move.l	(a7)+,RelocStackList-FirstRun(a6)	; Restore Old Stack List
	move.l	(a7)+,a5				; Restore Main Program Ref
	st.b	d0					; An error occured
	rts						;

Error_Table:
	dc.w	mmemory-Error_Table
	dc.w	mpastrouve-Error_Table
	dc.w	mwrong-Error_Table
	dc.w	wrongrom-Error_Table
	dc.w	wrongkernel-Error_Table
	dc.w	parityerror-Error_Table
	
; Standards relocs functions for nostub programs.
reloc:	movem.l	d1-d7/a1-a6,-(a7)
	moveq	#0,d7			; Relocation Mode selected
relocHd	move.w	d0,-(a7)		; Push handle
	ROM_THROW HeapDeref		; Get address
	addq.w	#2,a7			; Fix sp
	lea	2(a0),a5		; Program Ptr
relocDo	HW2TSR_PATCH a5,d0		; Patch if HW2 TSR
	bsr.s	start_relocation	; Relocation / Unrelocation
	move.l	ErrorString-FirstRun(a6),a0 ; Load the error string in a0 (Return value).
	movem.l	(a7)+,d1-d7/a1-a6	
	rts
reloc2:	movem.l	d1-d7/a1-a6,-(a7)
	moveq	#0,d7			; Relocation Mode selected
relocAn	move.l	a6,a5			; Get Program Ptr
	bra.s	relocDo			; Do relocation / unrelocation
unreloc	movem.l	d1-d7/a1-a6,-(a7)
	moveq	#1,d7			; UnRelocation Mode selected
	bra.s	relocHd			; UnRelocation of the selected handle
unreloc2 movem.l d1-d7/a1-a6,-(a7)
	moveq	#1,d7			; UnRelocation Mode selected
	bra.s	relocAn			; Unrelocation of the ptr



; Unreloc a program.
; In:
;	a5 -> Program.(HeapDeref(h)+2).
; Out:
;	a6 -> FirstRun (Preos Data Ptr)
;	d0.w = 0 if Ok, = -1 if an error occured while unrelocating (I don't think in normal utilisation it is possible).
;Destroy:
;	d7/a6/a1/d0
kernel::unrelocation:
	moveq	#1,d7			; Mode : UnRelocation
	bra.s	start_relocation


; Reloc a program.
; In:
;	a5 -> Program (HeapDeref(h)+2).
; Out:
;	a6 -> FirstRun (Preos Data Ptr)
;	d0.w = 0 if Ok, = -1 if an error occured while relocating.
;Destroy:
;	d7/a6/a1/d0
kernel::relocation:
	moveq	#0,d7				; Mode : Relocation

start_relocation:
	bsr	kernel_install_int		; Get a6 Ptr
	; How to get it from a Flash App ?
	pea	(a5)				; Push a5
	move.l	RelocStackList-FirstRun(a6),-(a7)	; Push old Stack Ptr (To keep the list) (I have choose this order because even vector NULL ptr is a valid entry -Well it will reset the calc but that's all-.
	move.l	a7,RelocStackList-FirstRun(a6)	; Save new Stack Frame Ptr
	bsr.s	relocation_recur		; relocation...
	move.l	(a7)+,RelocStackList-FirstRun(a6) ; Reload Old Ptr
	move.l	(a7)+,a5			; Reload Pc
	rts
	
; In:
;	a5 -> Program
;	a6 -> Data Ptr
;	d7.b = 0 if Reloc
;	d7.b = 1 if UnReloc
;	d7.w += $8000 if Non-recursive 
relocation_recur:
	movem.l d0-d7/a0-a6,-(a7)

	; Check internal format
	moveq	#4,d0				; Error: Wrong Kernel Version
	ifd	FLASH_CODE
		clr.w	KERNEL_EXTRA_NUMBER-FirstRun(a6)	; For Pedrom, we want to know what is the missing RamCall
		lea	Void_str(pc),a3
	endif
	tst.b	8(a5)
	bne	relocation::error
	
	; Read in d6.l the size of the file to avoid overflow while relocating
	moveq	#0,d6
	move.w	-2(a5),d6
	
	; Relocation counter
	tst.b	d7
	bne.s	\unrelloc
		ori.w	#$8000,-4(a5)		; Lock the handle 
		addq.b	#1,9(a5)		; Inc Reloc number
		cmp.b	#1,9(a5)		; CHeck if it is the first time
		bne 	fin_relocation		; We reloc this file
		bra.s	\cpt_relloc		; Yes, do not reloc it again. No, reloc it.
\unrelloc
		subq.b	#1,9(a5)
		bne 	fin_relocation		; Dec Reloc Number
\cpt_relloc
	
	; Set the file in-used.
	move.l	a5,a0
	bsr	kernel::Ptr2Hd
	bsr	kernel::Hd2Sym
	move.l	a0,d0
	beq.s	\Void
		bset.b	#0,SYM_ENTRY.flags(a0)
\Void:	

	lea	$1A(a5),a4			; a4 -> Import Table

	; Importation of the libraries
	move.w	(a4)+,d5				; d5.w = Number of used libraries
	beq.s	finish_lib				; If no library, skip Library relocation
		move.w	d5,d0				; Calcul the library offset table
		mulu.w	#10,d0				; d3=10*d5
		lea	0(a4,d0.w),a3			; a3 = List of the offset of the used librairies
		exg	a4,a3				; Exchange a4 & a3
\next_lib:		move.l	a3,a0			; a0 -> Searched Lib name
			move.b	9(a3),d0		; The program needs at least this version
			bsr	find_lib		; Find it (Unarchived / Uncompressed it also).
			move.l	a0,d0			; a0 = Address of the lib
			bne.s	\impect
				moveq	#1,d0			; "Lib not found"
				bra	relocation::error	; Go to error
\impect:		; Reloc the library
			tst.w	d7			; Check if we can do a recursevely relocation
			blt.s	\no_recur
				exg.l	a0,a5		; Relocation libs...
				bsr	relocation_recur
				exg.l	a0,a5		; The library has been relocated
\no_recur		move.l	a0,a1
			adda.w	$16(a1),a1		; Adress of the export table in a1
			move.w	(a1)+,d1		; Number of exported functions in d1
			move.w	(a4)+,d2		; Number of imported functions in d2
\reloc_lib			moveq	#0,d3		;
				move.w	(a4)+,d3	; Function Number in d3
				cmp.w	d1,d3		; Check if the function number is exported ?
				blt.s	\good_version
\wrong_version:				moveq	#2,d0	; The functions doesn't exist.
					bra	relocation::error
\good_version:			add.w	d3,d3		; Numberx2
				move.w	0(a1,d3.w),d3	; Function offset (from the library start) in d3
				move.l	a0,d4		; Adress of the lib in d4
				add.l	d3,d4		; Adress of the function in d4
				bsr	do_reloc	; Reloc the function
				dbf	d2,\reloc_lib	; Reloc all the imported functions
			lea	10(a3),a3		; Next library name.
			subq.w	#1,d5			; Library Counter -1
			bne.s	\next_lib		; If LibraryCounter>0, continue
finish_lib						; End of Libraries importation

	; RomCalls Importation
	move.w	(a4)+,d5				; Is there any Rom Calls ?
	beq.s	finish_rom				; No Rom Calls, go to RamCall
ROMBASE2	move.l	(ROM_VECTOR+$C8).l,a3		; ROM CALLS table 
		move.w (a4)+,d5				; Number of imported RomCalls in d5.w
\loop1			moveq	#0,d3			; Clear
			move.w	(a4)+,d3		; Function number in d3
			cmp.l	-4(a3),d3		; Check if the RomCall exists
			blt.s	\valid_romcall		; Yes, continue
\invalid_romcall		moveq	#3,d0		; No, Error: Wrong Rom Version
				ifd	PEDROM		; If Pedrom, we want to know
					move.w	-(a4),KERNEL_EXTRA_NUMBER-FirstRun(a6)	; What is the missing
					lea	Void_str(pc),a3			; romcall.
				endif
				bra	relocation::error
\valid_romcall		lsl.w	#2,d3			; d3*4 -> d3
			move.l	0(a3,d3.w),d4		; Adress of the exported romcall in d4
			ifd	PEDROM
				cmp.l	#INVALID_ROMCALL,d4	; Under Pedrom some RomCalls are not defined, but nervertheles exported
				beq.s	\invalid_romcall	; So I check if the RomCall address == INVALID_ROMCALL.
			endif
			bsr	do_reloc		; Reloc the RomCall
\endloop2	dbra	d5,\loop1			; Continue for all romcalls
finish_rom:						; End of RomCall importation

	;RAM relocation table
	move.w	(a4)+,d5				; Number of RAM_CALLS > 0 (But why do Xavier & Rusty have specify this word ???????)
	beq.s	finish_ram				; No RAMCALL, skip
		lea	RAM_TABLE(pc),a3		; a3 -> RamCall Table 
		move.w (a4)+,d5				; Number of different RAM_CALLS !
fonctionRAM:		moveq	#0,d3
			move.w	(a4)+,d3		; Read RAM_CALL number
			move.l	d3,d2			; Copy (because we work on RAM_CALL & EXTRA_RAM_CALL)
			lsl.w	#2,d3			; d2*4 <= Values == adress or longs words.
							; & Remove the word flag & the extra_ram flag
			btst	#14,d2			; Look for EXTRA_RAM_ADDRESS ?
			bne.s	extraramtable		; Yes => Do ExtraRamCall
				cmp.w	#MAX_RAMCALL*4-4,d3	; Check if valid RAM_CALL
				bls.s	validramcall	; Yes continue
					moveq	#4,d0	; Error: Wrong Kernel Version
					ifd	FLASH_CODE
						move.w	-(a4),KERNEL_EXTRA_NUMBER-FirstRun(a6)	; For Pedrom, we want to know what is the missing RamCall
						lea	Void_str(pc),a3
					endif
					bra	relocation::error
validramcall			move.l	0(a3,d3.w),d4		; valeur dans d4
				ifd	V200			; V200 only 
v200_92p_patch				tst.w	d3		; Calculator RamCall ? NOTE: This code will be erased by 'NOP' instruction by the installation code on 92+ & 89
					bne.s	loopRAMInit	; No so skip
						btst.b	#5,$11(a5)	; Check V200 flags for the program ?
						bne.s	loopRAMInit	; Yes do not falsify CALCULATOR
							addq.l	#4,d4	; Falsify the Value of CALCULATOR (because txtrider does some calculus on the values of CALCULATOR : it doesn't expect that CALCULATOR may be different from 0 or 1 !)
				endif
v200_92p_patch_end		bra.s	loopRAMInit
extraramtable:			add.w	$18(a5),d3		; For ExtraRamTable, add to the function offset, the offset of the EXTRA_RAM_TABLE
				moveq	#0,d4			; And 
ExtraRamCalc			move.w	IS89_0_2(a5,d3.l),d4	; Read the value for 89 or 92 (The installation program will the constante IS89_0_2 to 0 on 89 and to 2 on 92+/V200).
loopRAMInit		RELOC_MAC	d4			; d4 = d4 if reloc / d4 = ~d4 if unreloc.
\loopRAM:		moveq	#0,d3				;
			move.w	(a4)+,d3			; Read Offset 
			beq.s	\loopRAMend			; Offset = 0 => End of reloc table
				bsr	checkparity		; Check Overflow and parity of the offset.
				tst.w	d2			; Test bit 15 (Word / Long Word flag)
				blt.s	\word
					add.l	d4,0(a5,d3.l)	; Reloc a Long Word : add d4.l to the address
					bra.s	\loopRAM
\word					add.w	d4,0(a5,d3.l)	; Reloc a word : add d4.w to the address
					bra.s	\loopRAM
\loopRAMend	dbra	d5,fonctionRAM				; Continue for all the imported ramcalls
finish_ram							; End of RAM_CALL relocation

	;Relocation table (the global variables)
	move.l	a5,d4				; Adress of the address to reloc in d4 (The beginning of the program).
	bsr	do_reloc			; Relocation of the variables

	;Relocation of the BSS section
	move.w	$14(a5),d0			; Offset of the BSS table in d0
	beq.s	finish_BSS			; if =0, there is no BSS section.
		lea	0(a5,d0.w),a4		; Address of the BSS reloc table in a4
		tst.b	d7			; Test relocation ?
		bne.s	\unrelloc		; No go unrelloc
						; Relocation !
			move.l	(a4)+,-(a7)	; Push the size of the BSS block (Long so we can have BSS section > 64K with Pedrom)
			ROM_THROW HeapAllocHigh	; Alloc it as High as possible (and lock it)
			move.w	d0,$12(a5)	; Save the handle
			bne.s	\ok_mem		; Check if successfull
				moveq	#0,d0	; No, so an error occured
				bra	relocation::error
\ok_mem			move.w	d0,-(a7)	; Push the handle and 
			ROM_THROW HeapDeref	; Deref to get its adress
			HW2TSR_PATCH	a0,d0	; To prevent some problems if someone writes some code in the BSS section.
			move.l	a0,d4		; adress of the BSS block in d4
			clr.w	(a7)		; Fill with '0'
			pea	(a0)		; Push beginning address
			ROM_THROW memset	; Clear the allocated buffer
			lea	10(a7),a7	; Pop 10 (void *, Character, size)
			bra.s	\brarelloc	; Reloc the code
\unrelloc		move.w	$12(a5),-(a7)	; Save the handle of the BSS block
			bne.s	\HdExist	; 
				moveq	#0,d0	; No BSS handle => An error occured
				bra	relocation::error
\HdExist		ROM_THROW HeapDeref	; Deref it
			HW2TSR_PATCH	a0,d0	; To prevent some problems
			move.l	a0,d4		; Adress of the BSS block in d4
			addq.l	#4,a4		; Skip size of BSS block
			ROM_THROW HeapFree	; Free the handle
			clr.w	$12(a5)		; No more BSS Handle in the kernel header
			addq.l	#2,a7		; Fix stack ptr
\brarelloc
	bsr	do_reloc			; Relocation of the BSS section.
finish_BSS					; End of relocating the BSS

	; Test if we will free the temporary copy of the file.
	subq.w	#1,d7					; We do it only if it is a recusrsive unrealloc .
	bne.s	fin_relocation				; So we test if d7.w = 1
_reloc_free:	andi.w	#$7FFF,-4(a5)			; UnLock the handle
		;move.l	a5,a0				; a0 = a5
		;bsr	kernel::Ptr2Hd			; Calc Handle in d0 (2 octets en moins qu'avec HeapPtr... et on ne peut pas faire autrement si on utilise HW2TSR -ou un masque- mais ca marchera pas avec les libs archivees en RO !)
		;move.w	d0,d3				; Some big problems with Static Element of Pack Archive (They can be in 2 different handles !)
		;beq.s	fin_relocation			; No Handle ?
		; Why can not use kernel::Ptr2Hd ? Because a Static Pack Archive can be in 2 Handles ! So I must find the beginning of the Handle.
		move.l	a5,d3
		subq.l	#2,d3				; Get to the top
		bsr	kernel::PrepareHeapPtr		; Prepare the Ptr to avoid the ghost space (Due to HW2TSR)
		ifnd	PEDROM			; Preos uses the romcall HeapPtrToHandle
			move.l	d3,-(a7)		; Push the addr
			ROM_THROW HeapPtrToHandle	; Get the handle
			addq.l	#4,a7			
			move.w	d0,d3			; Error ?
			beq.s	fin_relocation		; Found ?
		endif
		ifd	PEDROM			; For Pedrom, we must rewrite the searching function since Pedrom's HeapPtrToHandle works more or less like Ptr2HD !
			move.w	#MAX_HANDLES-1,d2	; Max number of handles
			moveq	#-1,d0			; Handle
\loops				addq.w	#1,d0		; Handle ++
				cmp.l	(a1)+,d3	; Check Handle 
				dbeq	d2,\loops	; Continue ?
			bne.s	fin_relocation		; Not found.
			move.w	d0,d3				
		endif
			bsr	kernel::Hd2Sym		; Search in the VAT the handle of the current unrelocated library / program
			move.l	a0,d0			; Check if successfull
			beq.s	fin_relocation		; Enable to find the Handle in the VAT ?
			bclr.b	#0,SYM_ENTRY.flags(a0)	; Unused the file.
			move.l	a0,a2			; Because HeapFree destroy a0, and I need to get a SYM ptr
			btst.b	#6,SYM_ENTRY.flags(a0)	; Attribut = Local ?
			beq.s	\NoLocal
				addq.l	#SYM_ENTRY.compat,a0
				move.w	(a0),4(a0)	; Recopy Org archived Handle in hVal
				clr.w	(a0)+		; Clear Compat = 0
				bset	#1,(a0)		; Attribut = Archived
				andi.w	#~(64*256+16+8),(a0)	; Attribut = No Local, No Hidden, No lock
				move.w	d3,-(a7)	; d3 = Temp File Handle = Old SYM_ENTRY.hVal(a3)
				ROM_THROW HeapFree	; Free the copy : Free does not do any heap compression
				addq.l	#2,a7		; To do after modifing the SYM ?
\NoLocal		btst	#3,SYM_ENTRY.flags(a2)	; Attribut = Collapsed ? (Can not use twin else it crashes, because Tios assumes that if a var has the twin flags, then the next SYM_ENTRY is the original entry, and it will always skipped). And a normal file can not have these flag, so I can check if it has to be deleted !
			beq.s	fin_relocation		; A file can be Local & Collapsed which means it is a created archived file which have a temprorary copy in RAM... Yes, Preos even creates some archived files, and after needs to execute them, so it copies them in RAM - Too complicated for me-
				move.l	a2,a0		; a0 = SYM_ENTRY to delete
				bsr	SymDelPtr	; Delete symbol entry in the VAT.
fin_relocation						; End of relocating this file
	movem.l (a7)+,d0-d7/a0-a6			; Restore registers
	clr.w	d0					; Ok
return	rts						; Quit

reloc_free:
	movem.l d0-d7/a0-a6,-(a7)
	bra.s	_reloc_free

; In:
;	a4 -> Offset Table
;	d4.l = Value to add (or to sub)
;	d7 = Relocation / Unrelocation mode
;	a5 -> Program reference
;	d6 = Program Size
;Out:
;	Program relocated
;	a4 -> Next section
; Destroy:
;	d4 / d3 / a4
do_reloc:
	RELOC_MAC	d4			; If relocation, d4 =d4. If Unrelocation, d4 =~d4
	moveq	#0,d3
\loop:
	move.w	(a4)+,d3			; Offset in d3
	beq.s	return				; No offset ? => End of reloc function
		bsr.s	checkparity		; Check if no overflow and if offset is pair.
		add.l	d4,0(a5,d3.l)		; Reloc the address
		bra.s	\loop			; Continue
	
; Test of the parity of the offset and control program overflow
; In: 
;	Offset 		d3.l
;	Program Size	d6.l
; Out:
;	Nothing (An exepction is triggered if there is an error).
; Destroy:
;	Nothing
checkparity:
	cmp.l	d6,d3		; Test depassement du programme ?
	bge.s	\error
	btst.l	#0,d3		; Test parite de l'offset
	beq.s	return
\error		moveq	#5,d0	; Stub Error
		bra	relocation::error


; Function of clean-up of the Kernel programs.
; It searchs in the VAT all the kernel programs which are still
; relocated. If it find one, it :
;	+ Calls its exit-point.
;	+ Unrelocates it (without any recursivity).
;	+ Archiveds it / Frees it.
clean_up:
	; Find all files in the VAT
	move.w	#2,-(a7)			; Args for 
	clr.l	-(a7)				; SymFindFirst
	ROM_THROW SymFindFirst			; *and* stop flag pushed on the stack (0) !
\clean_loop					; Loop in VAT
		bsr	IsKernelPrgm		; Test if it is a Kernel program
		tst.w	d0			; Yes or No
		beq.s	\clean_next		; No so skip this file.
			move.w	SYM_ENTRY.hVal(a0),-(a7)	; Get its handle
			ROM_THROW HeapDeref	; Get its address
			addq.l	#2,a7		; Fix stack ptr
			addq.l	#2,a0		; Skip the File Size
			tst.b	9(a0)		; Reloc count != 0 ?
			beq.s	\clean_next	; No so skip it.
				pea	(a0)	; Push it on the stack for later treatement (This file is locked)
\clean_next	ROM_THROW SymFindNext		; Next file. Can block if some AMS global vars are corrupted (Solution ON+ESC).
		move.l	a0,d0			; Check if we have another SYM_ENTRY ?
		bne.s	\clean_loop		; Yes, quit
	move.l	a7,a4				; Save the tab address (Top of the stack).
	; First, calls all the exit points
\loopexit	move.l	(a4)+,d0		; Get the file
		beq.s	\loopexitend		; End of table, quits.
			move.l	d0,a0		; Get the address of the file
			moveq	#0,d0
			move.w	$E(a0),d0	; Get the exit offset
			beq.s	\no_exit_point	; No exit function, skip it.
				cmpi.b	#$FF,$9(a0)		; Test if we have already called this exit function ?
				beq.s	\no_exit_point		; Yes, so it crashes. Do not call it again !
					st.b	$9(a0)		; Protection against infinite call to exit point
					pea	(a4)		; Push a4
					jsr	0(a0,d0.l)	; Call exit point
					move.l	(a7)+,a4	; Pop a4
					bsr	kernel_install_int	; Reinstall a6 / Vectors
\no_exit_point:		bra.s	\loopexit	; Continue for all the relocated kernel files.
\loopexitend
	move.l	a7,a4					; Restore file tab
\loopunrelloc	move.l	(a4)+,d0			; Get the file ptr
		beq.s	\loopunrellocend		; End of table ?
			move.l	d0,a5			; Get the address
			move.b	#1,$9(a5)		; Reloc count = 1, so that after unreloc it will be 0
			pea	(a4)			; Push register
			move.w	#$8000+1,d7		; Mode Unreallocation without recursivity
			bsr	relocation_recur	; Unreloc the file.
			move.l	(a7)+,a4		; Reget a4
			bra.s	\loopunrelloc		; Continue
\loopunrellocend		
	; Free the archived copies (and anything else)
\loopfree	move.l	(a7)+,d0			; Get the file (Do it with a7 so that we pop the stack too)
		beq.s	\loopend			; End of table ?
			move.l	d0,a5			; Get the address
			bsr	reloc_free		; Free the copy if it exists
			bra.s	\loopfree		; Continue
\loopend
	addq.w	#2,a7					; Fix the stack
	rts

; Check if it is a Kernel Program.
; In:
;	a0.l -> Sym Entry
; Out:
;	d0 = 0 if it isn't a kernel program.
;	d6.b = 0
IsKernelPrgm:
	clr.b	d6
	
; Check if it is a Kernel Program with a good version.
; In:
;	a0.l -> Sym Entry
;	d6.b = Min Library Version required
; Out:
;	d0 = 0 if it isn't a kernel program.
IsKernelVersionPrgm:
	movem.l	d1-d2/a0-a1,-(a7)
	move.w	SYM_ENTRY.hVal(a0),-(a7)		; Get Handle and pushs it on the stack
	ROM_THROW HeapDeref				; Deref the handle and get its address in a0
	clr.w	d0					; Default: Fail
	tst.w	(a7)+					; In case, there is no handle for the var (Local vars not init for example).
	beq.s	\echec					; So it fails.		
		moveq	#0,d1				; Clear d1.l
		move.w	(a0)+,d1			; To read file size (File size is unsigned short).
		cmp.b	#$F3,-1(a0,d1.l)		; Check ASM TAG ?
		bne.s	\echec				; No, it isn't an ASM file.
			move.l	4(a0),d1		; Read Magic Signature
			lsr.l	#8,d1			; 
			cmpi.l	#'68k',d1		; It must me be '68k?' where ? is a joker.
			bne.s	\echec			; 
				cmp.b	VERSION_OFFSET(a0),d6	; Check version number
				sls.b	d0			; Set d0 according to the result of the test
\echec	movem.l	(a7)+,d1-d2/a0-a1
	rts
	

; Find the missing library in the VAT.
; Unarchive it / Uncompress it / Unpack it if necessary
; In:
;	a0 ->  Lib Name
;	d0.b = Min Library Version required
; Out:
;	a0 -> Unarchived Lib (First line of code) or NULL.
find_lib:
	movem.l	d0-d6/a1-a4,-(a7)
	
	move.b	d0,d6					; Save version
	move.l	a0,a2					; Save Name

	move.w	#2,-(a7)				; VAT loop
	clr.l	-(a7)					; Parameters for SymFindFirst : All VAT
	ROM_THROW SymFindFirst				; Get first Sym Ptr
\loop		; Check if it a Kernel Asm Program, check for the name, and the library version
		move.l	a0,a3				; Save SYM ptr 
		bsr.s	IsKernelVersionPrgm		; First Check if it is a Kernel Program ?
		tst.w	d0				; with the good version number, too !
		beq.s	\no_kernel			; No kernel Program => No lib file.
			pea	(a0)			; Push found sym
			pea	(a2)			; Push searched file
			ROM_THROW SymCmp		; Compare the 2 files.
			addq.l	#8,a7			; Fix the stack
			tst.w	d0			; Is it the good file ?
			bne.s	\no_kernel		; No, next file, please.
\ker_found			move.l	a3,a0		; Arg in a0
				bsr	kernel::Ptr2Hd	; Get Folder Handle of SYM_ENTRY
				pea	(a3)		; Push SYM (2nd arg)
				move.w	d0,-(a7)	; Push Folder Handle (1st arg)
				ROM_THROW MakeHsym	; Call Function
				addq.l	#6,a7		; Pop args
				move.l	d0,d3		; Save HSym in d3
				move.w	SYM_ENTRY.hVal(a3),(a7); We have found it.
				ROM_THROW HeapDeref	; Deref the handle.
				btst.b	#3,($11+2)(a0)	; Test the Read Only flag.
				bne.s	\find2		; If ro set, do not unarchive it ! and do not add Hw2Tsr patch !
				cmp.l	#$200000,a0	; Check Archive Mem ? (I don't test the flag because it may be called with PackArchive Static Handle.)
				blt.s	\find		; Archived ? ; Archive => Create a temporary copy.
					move.l	a0,a2		; Save the adress of the lib
					moveq	#3,d5		; Size + 3 to get the real size (FIXME : only +2 ?)
					add.w	(a2),d5		; d5 = File size
					move.l	d5,(a7)		; Push File Size
					ROM_THROW HeapAllocHigh	; And alloc a new Handle for the copy.
					move.l	d3,(a7)		; Push HSym
					move.w	d0,d3		; Check if HeapAlloc successfull 
					beq.s	\no_kernel	; No, so we will search for another file (FIXME: is it a good idea ?)
					ROM_THROW DerefSym
					move.w	SYM_ENTRY.hVal(a0),SYM_ENTRY.compat(a0)	; Save org handle in compat field (ASM file are not compatible with ti-92)
					move.w	d3,SYM_ENTRY.hVal(a0)			; It works fine (Fixme: if there are many handles - >256 -, does it work very fine ?)
					ori.w	#64*256+16+8,SYM_ENTRY.flags(a0)	; Attribut = Local & Hidden & Lock
					bclr	#1,SYM_ENTRY.flags(a0)			; Attribut = unarchived
					move.w	d3,(a7)		; Repush the handle
					ROM_THROW HeapDeref	; to rederef it
					move.l	a0,a1		; Recopy the file
\copy_loop:					move.b	(a2)+,(a1)+
						subq.w	#1,d5		; Better than subq+dbf (6 bytes)
						bne.s	\copy_loop	; subq + bne = 4 bytes (but slower)
\find				HW2TSR_PATCH a0,d0		; Add +$40000 to enter the ghost space if Hw2Tsr
\find2				addq.l	#2,a0			; Skip the File Size
				bra.s	\end			; Found !
\no_kernel	ROM_THROW SymFindNext			; Next File
		move.l	a0,d0				; Check end of VAT ?
		bne	\loop				; No => continue
	; The library has not been found in the VAT.
	bsr	kernel::ExtractFile			; Search a compressed file with this name.
	move.w	d0,(a7)					; Success ? Handle != 0?
	beq.s	\ferror					; Error
		ROM_THROW HeapLock			; Lock Handle
		lea	-20(a7),a7			; Transform it in a SYM STR
		move.l	a7,a1				; To be compatible
		clr.b	(a1)+				; with AMS file name convention
\cpy			move.b	(a2)+,(a1)+		; Recopy it between 2 '0'
			bne.s	\cpy
		pea	-1(a1)				; Push var name
		ROM_THROW SymFindMain			; Test if filename is in the VAT (To prevent some problem)
		tst.l	d0				; Test if we find it ?
		bne.s	\terror				; Yes => Error (Free the handle and fix the stack)
		ROM_THROW SymAddMain			; Add var in main folder.
		move.l	d0,(a7)				; And check for it
		ROM_THROW DerefSym			; Deref from HSym struct to SYM_ENTRY * ptr
		move.l	a0,d0				; Check if SYM_ENTRY * != 0
		bne.s	\done				; No error, continue
\terror			lea	24(a7),a7		; Fix the stack (due to the TIOS name)
			ROM_THROW HeapFree		; Free the uncompressed handle since we cannot add it in the VAT
			bra.s	\ferror			; FIXME: Was it tested successfully ?
\done		lea	24(a7),a7			; Fix the stack (due to the TIOS name)
		move.w	(a7),SYM_ENTRY.hVal(a0)		; Save the handle in the created SYM
		ori.w	#8*256+16+8,SYM_ENTRY.flags(a0)	; Attribut = Collapsed, Hidden, locked (Yes a file never could have theses flags normally so I can check it :p)
		move.l	a0,a3				; Is it a kernel program with the good version ?
		bsr	IsKernelVersionPrgm		; FIXME: Maybe I should test it before ?
		tst.w	d0				
		bne	\ker_found			; Kernel => Unarchive it if necessary
			bsr.s	SymDelPtr		; No so del the file
\ferror			suba.l	a0,a0			; Error: Return 0
\end	addq.l	#6,a7					; Fix the stack
	movem.l	(a7)+,d0-d6/a1-a4			;
	rts						; Return in a0 the lib ptr or NULL.

; Delete the symbol a0
; In:
;	a0 -> SYM_ENTRY
SymDelPtr:
	clr.w	SYM_ENTRY.flags(a0)		; Clear the flags (Mainly in-use flag)
	pea	(a0)				; Push SYM_ENTRY
	bsr	kernel::Ptr2Hd			; Get the handle of the folder
	move.w	d0,-(a7)			; Push Folder Handle
	ROM_THROW MakeHsym			; Create the HSym
	move.l	d0,(a7)				; Push HSym
	ROM_THROW HSymDel			; Delete it
	addq.l	#6,a7				; Pop the stack
	rts
	